#! eggdrop
# $Id: usertopic.tcl,v 1.0 2014/12/10 13:50:52 eggderp Rel eggderp $
#
# Allows any user to change the topic after the last | character so
# non-opped users can change part of the topic. | characters in this topic are
# replaced with a different character so that non-op users can't shift the end
# of the user-modifiable topic.
#
# Optionally, the script can also log all topic changes to an SQLite database
# so previous topics can be displayed on a website using a CGI script written
# in Tcl (See www directory).
#
# To enable, after configuring the settings below, use
#  .chanset #example +usertopic
# to allow non-opped users to change the last part of the topic and - if sqlite
# is enabled - use
#  .chanset #example +logtopics
# to log all topic changes in #example to the sqlite database.

# Copyright (c) 2014, Moritz Wilhelmy (ente on IRCnet)
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

###########################################################################
#### USER CONFIGURABLE SETTINGS, EDIT THESE BEFORE ENABLING THE SCRIPT ####
###########################################################################

set prefix "!"  ;# the character your channel's triggers start with
namespace eval "usertopic" {
variable settings;
set settings(ignore-flags)        "bdkqr|dkqr" ;# ignore users with these flags set
set settings(udef-flag)           "usertopic"  ;# name of the udef-flag to use
set settings(log-topics)          "yes"        ;# log topics to a SQLite database file?
set settings(sqlite-db)           "topics.db"  ;# filename of the SQLite database
set settings(udef-log)            "logtopics"  ;# flag to decide whether to log topics on a channel
set settings(topic-url)           "http://example.com/topics.cgi" ;# address where people can find the full log

###########################################################################
### END OF USER SETTINGS. ONLY EDIT BELOW IF YOU KNOW WHAT YOU'RE DOING ###
###########################################################################

set skip_sqlite3 0 ;# while messing with the database, .set this to 1, then reload the script

variable VERSION {$Revision: 1.0 $}
set VERSION [regsub {\$[R]evision: (.+) \$} $VERSION {\1}]

variable ns [namespace current]

setudef flag $settings(udef-flag)

proc may_set_topic {nick hand chan} {
	variable settings
	# Add conditions here if you need to.
	if {[matchattr $hand $settings(ignore-flags)]} { return 0 }
	if {[isop $nick $chan]} { return 1 }
	if {![channel get $chan $settings(udef-flag)]} { return 0 }
	return 1
}

proc set_topic {nick uhost hand chan text} {
	if {![may_set_topic $nick $hand $chan]} return

	set oldtopic [topic $chan]

	if {$text == ""} {
		# if no new topic was given, broadcast the old one for good measure
		# can be useful for e.g. IRCnet where topics sometimes get lost on netsplit.
		# don't log it since it wasn't modified.
		puthelp "TOPIC $chan :$oldtopic"
		return
	}

	set topic [lreplace [split $oldtopic |] end end] ;# remove last element
	set text [string map {| ¦} $text];# replace all | so the user doesn't introduce a new unmodifiable element
	lappend topic $text

	set newtopic {} ;# trim spaces from each element
	foreach el $topic {
		lappend newtopic [string trim $el]
	}

	set newtopic [join $newtopic " | "]
	if {$newtopic eq $oldtopic} {
		puthelp "PRIVMSG $chan :Topic not modified."
		return
	}
	puthelp "TOPIC $chan :$newtopic"
	log_topic $nick $uhost $hand $chan $newtopic ;# fake the log entry so that the user rather than the bot appears in the log 
	return
}

bind pub * "${prefix}topic" ${ns}::set_topic

##### Topic Logging #################
package require sqlite3
setudef flag $settings(udef-log)

proc reopen_db { } {
	# this procedure must be manually invoked via .tcl usertopic::reopen_db if the database is moved
	# while messing with the database, use .set usertopic::skip_sqlite3 1 to disable writes to the database until you're done.
	variable skip_sqlite3;
	variable settings;
	variable ns;

	if {! [string length $settings(sqlite-db)]} {
		set skip_sqlite3 1
	}

	if {$skip_sqlite3} return;

	if {[llength [info commands ${ns}::db]]} {
		db close
	}

	sqlite3 ${ns}::db $settings(sqlite-db)
	init_db
}

proc init_db {} {
	variable skip_sqlite3;
	if {$skip_sqlite3} return
	db eval {
		CREATE TABLE IF NOT EXISTS
			topics( seq            INTEGER PRIMARY KEY  -- sequence number
			      , topic          TEXT NOT NULL
			      , chan           TEXT NOT NULL
			      , ts             INTEGER NOT NULL
			      , nick           TEXT NOT NULL
			      , egg_handle     TEXT
			      , uhost          TEXT NOT NULL
			);
	}
}

proc total {chan} {
	db eval {
		SELECT COUNT(*) AS total FROM topics WHERE chan = :chan
	} { return $total }
	return 0
}

proc sql_enabled {chan} {
	variable settings
	variable skip_sqlite3
	if {$skip_sqlite3} {return 0}
	if {[channel get $chan $settings(udef-log)]} {return 1}
	return 0
}


proc show_topics {nick uhost hand chan txt} {
	variable settings

	if {![sql_enabled $chan]} return

	putmsg $chan "Showing previous 3 topics out of [total $chan] for $chan (Use !revert <id> to go back, full list: $settings(topic-url)?channel=[string map {# %23} $chan] ):"
	db eval { SELECT
			  seq
			, COALESCE(egg_handle, nick) AS user
			, ts
			, topic
		FROM topics
		WHERE chan = :chan
		ORDER BY ts DESC
		LIMIT 3
		OFFSET 1
	} {
		set time [clock format $ts -gmt 1 -format "%Y-%m-%d %H:%M:%S"]
		putmsg $chan "\[$seq\] $time by \002$user\002: $topic"
	}
	return 1
}

proc log_topic {nick uhost hand chan topic} {
	global botname

	if {![sql_enabled $chan]} return
	if {"$nick!$uhost" eq $botname} return ;# ignore topic changes by the bot

	if {$nick eq "*"} {
		# bot joined the channel and got that message
		return 0
	}
	
	if {$hand eq "*"} {
		unset hand ;# hack to get sqlite to insert a NULL in this case
	}

	db eval {
		INSERT INTO topics(topic, chan, ts, nick, egg_handle, uhost)
			VALUES( :topic
			      , :chan
			      , strftime('%s', 'now')
			      , :nick
			      , :hand
			      , :uhost
			);
	}
	return 1
}

proc revert_topic {nick uhost hand chan txt} {
	if {![sql_enabled $chan] || ![may_set_topic $nick $hand $chan]} return

	set txt [string trim $txt]
	set oldtopic [topic $chan]

	set sql {SELECT topic FROM topics WHERE seq = :txt AND chan = :chan}

	if {$txt eq ""} {
		set sql {SELECT topic FROM topics WHERE chan = :chan ORDER BY seq DESC LIMIT 1 OFFSET 1}
	}

	db eval $sql {
		if {$topic eq $oldtopic} {
			puthelp "PRIVMSG $chan :Topic not modified."
			return 0
		}
		puthelp "TOPIC $chan :$topic"
		log_topic $nick $uhost $hand $chan $topic ;# fake it again
		return 1
	}
	
	putmsg $chan "No such topic."
	return 0
}

bind pub - ${prefix}topics ${ns}::show_topics
bind pub - ${prefix}revert ${ns}::revert_topic
bind topc - * ${ns}::log_topic

reopen_db
putlog "usertopic $VERSION loaded"
}
